class CatalogError(Exception):
    pass


class JSONPatchError(Exception):
    pass


class JSONPointerError(Exception):
    pass


class RelativeJSONPointerError(Exception):
    pass


class JSONSchemaError(Exception):
    pass


class URIError(Exception):
    pass
